@@ -6,7 +6,6 @@ from __future__ import annotations
 
 import asyncio
 import inspect
-import itertools
 import logging
 import os
 import socket
@@ -16,11 +15,13 @@ import time
 import typing as t
 import uuid
 import warnings
+from collections.abc import Mapping
+from contextvars import Context, ContextVar, copy_context
 from datetime import datetime
 from functools import partial
 from signal import SIGINT, SIGTERM, Signals, default_int_handler, signal
 
-from .control import CONTROL_THREAD_NAME
+from .thread import CONTROL_THREAD_NAME
 
 if sys.platform != "win32":
     from signal import SIGKILL
@@ -40,7 +41,6 @@ import zmq
 from IPython.core.error import StdinNotImplementedError
 from jupyter_client.session import Session
 from tornado import ioloop
-from tornado.queues import Queue, QueueEmpty
 from traitlets.config.configurable import SingletonConfigurable
 from traitlets.traitlets import (
     Any,
@@ -50,7 +50,6 @@ from traitlets.traitlets import (
     Instance,
     Integer,
     List,
-    Set,
     Unicode,
     default,
     observe,
@@ -61,11 +60,22 @@ from ipykernel.jsonutil import json_clean
 
 from ._version import kernel_protocol_version
 from .iostream import OutStream
+from .utils import LazyDict
+
+_AWAITABLE_MESSAGE: str = (
+    "For consistency across implementations, it is recommended that `{func_name}`"
+    " either be a coroutine function (`async def`) or return an awaitable object"
+    " (like an `asyncio.Future`). It might become a requirement in the future."
+    " Coroutine functions and awaitables have been supported since"
+    " ipykernel 6.0 (2021). {target} does not seem to return an awaitable"
+)
+
+T = t.TypeVar("T")
 
 
 def _accepts_parameters(meth, param_names):
     parameters = inspect.signature(meth).parameters
-    accepts = {param: False for param in param_names}
+    accepts = dict.fromkeys(param_names, False)
 
     for param in param_names:
         param_spec = parameters.get(param)
@@ -144,6 +154,7 @@ class Kernel(SingletonConfigurable):
     debug_shell_socket = Any()
 
     control_thread = Any()
+    shell_channel_thread = Any()
     iopub_socket = Any()
     iopub_thread = Any()
     stdin_socket = Any()
@@ -186,8 +197,17 @@ class Kernel(SingletonConfigurable):
 
     # track associations with current request
     _allow_stdin = Bool(False)
-    _parents: Dict[str, t.Any] = Dict({"shell": {}, "control": {}})
-    _parent_ident = Dict({"shell": b"", "control": b""})
+    _control_parent: Dict[str, t.Any] = Dict({})
+    _control_parent_ident: bytes = b""
+    _shell_parent: ContextVar[dict[str, Any]]
+    _shell_parent_ident: ContextVar[bytes]
+    _shell_context: Context
+    # Kept for backward-compatibility, accesses _control_parent_ident and _shell_parent_ident,
+    # see https://github.com/jupyterlab/jupyterlab/issues/17785
+    _parent_ident: Mapping[str, bytes]
+
+    # Asyncio lock for main shell thread.
+    _main_asyncio_lock: asyncio.Lock
 
     @property
     def _parent_header(self):
@@ -239,9 +259,6 @@ class Kernel(SingletonConfigurable):
     # by record_ports and used by connect_request.
     _recorded_ports = Dict()
 
-    # set of aborted msg_ids
-    aborted = Set()
-
     # Track execution count here. For IPython, we override this to use the
     # execution count we store in the shell.
     execution_count = 0
@@ -260,16 +277,17 @@ class Kernel(SingletonConfigurable):
         "shutdown_request",
         "is_complete_request",
         "interrupt_request",
-        # deprecated:
-        "apply_request",
     ]
-    # add deprecated ipyparallel control messages
+
+    # control channel accepts all shell messages
+    # and some of its own
     control_msg_types = [
         *msg_types,
-        "clear_request",
-        "abort_request",
         "debug_request",
         "usage_request",
+        "create_subshell_request",
+        "delete_subshell_request",
+        "list_subshell_request",
     ]
 
     def __init__(self, **kwargs):
@@ -296,7 +314,28 @@ class Kernel(SingletonConfigurable):
             self.do_execute, ["cell_meta", "cell_id"]
         )
 
+        self._control_parent = {}
+        self._control_parent_ident = b""
+
+        self._shell_parent = ContextVar("shell_parent")
+        self._shell_parent.set({})
+        self._shell_parent_ident = ContextVar("shell_parent_ident")
+        self._shell_parent_ident.set(b"")
+        self._shell_context = copy_context()
+
+        # For backward compatibility so that _parent_ident["shell"] and _parent_ident["control"]
+        # work as they used to for ipykernel >= 7
+        self._parent_ident = LazyDict(
+            {
+                "control": lambda: self._control_parent_ident,
+                "shell": lambda: self._get_shell_context_var(self._shell_parent_ident),
+            }
+        )
+
+        self._main_asyncio_lock = asyncio.Lock()
+
     async def dispatch_control(self, msg):
+        """Dispatch a control request, ensuring only one message is processed at a time."""
         # Ensure only one control message is processed at a time
         async with self._control_lock:
             await self.process_control(msg)
@@ -342,20 +381,35 @@ class Kernel(SingletonConfigurable):
         """Check whether a shell-channel message should be handled
 
         Allows subclasses to prevent handling of certain messages (e.g. aborted requests).
+
+        .. versionchanged:: 7
+            Subclass should_handle _may_ be async.
+            Base class implementation is not async.
         """
-        msg_id = msg["header"]["msg_id"]
-        if msg_id in self.aborted:
-            # is it safe to assume a msg_id will not be resubmitted?
-            self.aborted.remove(msg_id)
-            self._send_abort_reply(stream, msg, idents)
-            return False
         return True
 
-    async def dispatch_shell(self, msg):
+    async def dispatch_shell(self, msg, /, subshell_id: str | None = None):
         """dispatch shell requests"""
+        if len(msg) == 1 and msg[0].buffer == b"stop aborting":
+            # Dummy "stop aborting" message to stop aborting execute requests on this subshell.
+            # This dummy message implementation allows the subshell to abort messages that are
+            # already queued in the zmq sockets/streams without having to know any of their
+            # details in advance.
+            if subshell_id is None:
+                self._aborting = False
+            else:
+                self.shell_channel_thread.manager.set_subshell_aborting(subshell_id, False)
+            return
+
         if not self.session:
             return
 
+        if self._supports_kernel_subshells:
+            assert threading.current_thread() not in (
+                self.control_thread,
+                self.shell_channel_thread,
+            )
+
         idents, msg = self.session.feed_identities(msg, copy=False)
         try:
             msg = self.session.deserialize(msg, content=True, copy=False)
@@ -368,12 +422,25 @@ class Kernel(SingletonConfigurable):
         self._publish_status("busy", "shell")
 
         msg_type = msg["header"]["msg_type"]
+        assert msg["header"].get("subshell_id") == subshell_id
+
+        if self._supports_kernel_subshells:
+            stream = self.shell_channel_thread.manager.get_subshell_to_shell_channel_socket(
+                subshell_id
+            )
+        else:
+            stream = self.shell_stream
 
         # Only abort execute requests
-        if self._aborting and msg_type == "execute_request":
-            self._send_abort_reply(self.shell_stream, msg, idents)
-            self._publish_status_and_flush("idle", "shell", self.shell_stream)
-            return
+        if msg_type == "execute_request":
+            if subshell_id is None:
+                aborting = self._aborting  # type:ignore[unreachable]
+            else:
+                aborting = self.shell_channel_thread.manager.get_subshell_aborting(subshell_id)
+            if aborting:
+                self._send_abort_reply(stream, msg, idents)
+                self._publish_status_and_flush("idle", "shell", stream)
+                return
 
         # Print some info about this message and leave a '--->' marker, so it's
         # easier to trace visually the message chain when debugging.  Each
@@ -381,8 +448,12 @@ class Kernel(SingletonConfigurable):
         self.log.debug("\n*** MESSAGE TYPE:%s***", msg_type)
         self.log.debug("   Content: %s\n   --->\n   ", msg["content"])
 
-        if not self.should_handle(self.shell_stream, msg, idents):
-            self._publish_status_and_flush("idle", "shell", self.shell_stream)
+        should_handle: bool | t.Awaitable[bool] = self.should_handle(stream, msg, idents)
+        if inspect.isawaitable(should_handle):
+            should_handle = await should_handle
+        if not should_handle:
+            self._publish_status_and_flush("idle", "shell", stream)
+            self.log.debug("Not handling %s:%s", msg_type, msg["header"].get("msg_id"))
             return
 
         handler = self.shell_handlers.get(msg_type, None)
@@ -395,7 +466,7 @@ class Kernel(SingletonConfigurable):
             except Exception:
                 self.log.debug("Unable to signal in pre_handler_hook:", exc_info=True)
             try:
-                result = handler(self.shell_stream, idents, msg)
+                result = handler(stream, idents, msg)
                 if inspect.isawaitable(result):
                     await result
             except Exception:
@@ -413,7 +484,7 @@ class Kernel(SingletonConfigurable):
             sys.stdout.flush()
         if sys.stderr is not None:
             sys.stderr.flush()
-        self._publish_status_and_flush("idle", "shell", self.shell_stream)
+        self._publish_status_and_flush("idle", "shell", stream)
 
     def pre_handler_hook(self):
         """Hook to execute before calling message handler"""
@@ -438,11 +509,6 @@ class Kernel(SingletonConfigurable):
             if self.eventloop is not eventloop:
                 self.log.info("exiting eventloop %s", eventloop)
                 return
-            if self.msg_queue.qsize():
-                self.log.debug("Delaying eventloop due to waiting messages")
-                # still messages to process, make the eventloop wait
-                schedule_next()
-                return
             self.log.debug("Advancing eventloop %s", eventloop)
             try:
                 eventloop(self)
@@ -461,88 +527,11 @@ class Kernel(SingletonConfigurable):
             # already consumed from the queue by process_one and the queue is
             # technically empty.
             self.log.debug("Scheduling eventloop advance")
-            self.io_loop.call_later(0.001, partial(self.schedule_dispatch, advance_eventloop))
+            self.io_loop.call_later(0.001, advance_eventloop)
 
         # begin polling the eventloop
         schedule_next()
 
-    async def do_one_iteration(self):
-        """Process a single shell message
-
-        Any pending control messages will be flushed as well
-
-        .. versionchanged:: 5
-            This is now a coroutine
-        """
-        # flush messages off of shell stream into the message queue
-        if self.shell_stream:
-            self.shell_stream.flush()
-        # process at most one shell message per iteration
-        await self.process_one(wait=False)
-
-    async def process_one(self, wait=True):
-        """Process one request
-
-        Returns None if no message was handled.
-        """
-        if wait:
-            t, dispatch, args = await self.msg_queue.get()
-        else:
-            try:
-                t, dispatch, args = self.msg_queue.get_nowait()
-            except (asyncio.QueueEmpty, QueueEmpty):
-                return
-
-        if self.control_thread is None and self.control_stream is not None:
-            # If there isn't a separate control thread then this main thread handles both shell
-            # and control messages. Before processing a shell message we need to flush all control
-            # messages and allow them all to be processed.
-            await asyncio.sleep(0)
-            self.control_stream.flush()
-
-            socket = self.control_stream.socket
-            while socket.poll(1):
-                await asyncio.sleep(0)
-                self.control_stream.flush()
-
-        await dispatch(*args)
-
-    async def dispatch_queue(self):
-        """Coroutine to preserve order of message handling
-
-        Ensures that only one message is processing at a time,
-        even when the handler is async
-        """
-
-        while True:
-            try:
-                await self.process_one()
-            except Exception:
-                self.log.exception("Error in message handler")
-
-    _message_counter = Any(
-        help="""Monotonic counter of messages
-        """,
-    )
-
-    @default("_message_counter")
-    def _message_counter_default(self):
-        return itertools.count()
-
-    def schedule_dispatch(self, dispatch, *args):
-        """schedule a message for dispatch"""
-        idx = next(self._message_counter)
-
-        self.msg_queue.put_nowait(
-            (
-                idx,
-                dispatch,
-                args,
-            )
-        )
-        # ensure the eventloop wakes up
-        self.io_loop.add_callback(lambda: None)
-
     async def _create_control_lock(self):
         # This can be removed when minimum python increases to 3.10
         self._control_lock = asyncio.Lock()
@@ -550,8 +539,6 @@ class Kernel(SingletonConfigurable):
     def start(self):
         """register dispatchers for streams"""
         self.io_loop = ioloop.IOLoop.current()
-        self.msg_queue: Queue[t.Any] = Queue()
-        self.io_loop.add_callback(self.dispatch_queue)
 
         if self.control_stream:
             self.control_stream.on_recv(self.dispatch_control, copy=False)
@@ -565,17 +552,67 @@ class Kernel(SingletonConfigurable):
             self._control_lock = asyncio.Lock()
 
         if self.shell_stream:
-            self.shell_stream.on_recv(
-                partial(
-                    self.schedule_dispatch,
-                    self.dispatch_shell,
-                ),
-                copy=False,
-            )
+            if self.shell_channel_thread:
+                self.shell_channel_thread.manager.set_on_recv_callback(self.shell_main)
+                self.shell_stream.on_recv(self.shell_channel_thread_main, copy=False)
+            else:
+                self.shell_stream.on_recv(
+                    partial(self.shell_main, None),
+                    copy=False,
+                )
 
         # publish idle status
         self._publish_status("starting", "shell")
 
+    async def shell_channel_thread_main(self, msg):
+        """Handler for shell messages received on shell_channel_thread"""
+        assert threading.current_thread() == self.shell_channel_thread
+
+        async with self.shell_channel_thread.asyncio_lock:
+            if self.session is None:
+                return
+
+            # deserialize only the header to get subshell_id
+            # Keep original message to send to subshell_id unmodified.
+            _, msg2 = self.session.feed_identities(msg, copy=False)
+            try:
+                msg3 = self.session.deserialize(msg2, content=False, copy=False)
+                subshell_id = msg3["header"].get("subshell_id")
+
+                # Find inproc pair socket to use to send message to correct subshell.
+                subshell_manager = self.shell_channel_thread.manager
+                socket = subshell_manager.get_shell_channel_to_subshell_socket(subshell_id)
+                assert socket is not None
+                socket.send_multipart(msg, copy=False)
+            except Exception:
+                self.log.error("Invalid message", exc_info=True)  # noqa: G201
+
+    async def shell_main(self, subshell_id: str | None, msg):
+        """Handler of shell messages for a single subshell"""
+        if self._supports_kernel_subshells:
+            if subshell_id is None:
+                assert threading.current_thread() == self.shell_channel_thread.parent_thread
+                asyncio_lock = self._main_asyncio_lock
+            else:
+                assert threading.current_thread() not in (
+                    self.shell_channel_thread,
+                    self.shell_channel_thread.parent_thread,
+                )
+                asyncio_lock = self.shell_channel_thread.manager.get_subshell_asyncio_lock(
+                    subshell_id
+                )
+        else:
+            assert subshell_id is None
+            asyncio_lock = self._main_asyncio_lock
+
+        # Whilst executing a shell message, do not accept any other shell messages on the
+        # same subshell, so that cells are run sequentially. Without this we can run multiple
+        # async cells at the same time which would be a nice feature to have but is an API
+        # change.
+        assert asyncio_lock is not None
+        async with asyncio_lock:
+            await self.dispatch_shell(msg, subshell_id=subshell_id)
+
     def record_ports(self, ports):
         """Record the ports that this kernel is using.
 
@@ -615,7 +652,7 @@ class Kernel(SingletonConfigurable):
     def _publish_status_and_flush(self, status, channel, stream, parent=None):
         """send status on IOPub and flush specified stream to ensure reply is sent before handling the next reply"""
         self._publish_status(status, channel, parent)
-        if stream:
+        if stream and hasattr(stream, "flush") and not self._supports_kernel_subshells:
             stream.flush(zmq.POLLOUT)
 
     def _publish_debug_event(self, event):
@@ -638,8 +675,14 @@ class Kernel(SingletonConfigurable):
         The parent identity is used to route input_request messages
         on the stdin channel.
         """
-        self._parent_ident[channel] = ident
-        self._parents[channel] = parent
+        if channel == "control":
+            self._control_parent_ident = ident
+            self._control_parent = parent
+        else:
+            self._shell_parent_ident.set(ident)
+            self._shell_parent.set(parent)
+            # preserve the last call to set_parent
+            self._shell_context = copy_context()
 
     def get_parent(self, channel=None):
         """Get the parent request associated with a channel.
@@ -664,7 +707,22 @@ class Kernel(SingletonConfigurable):
             else:
                 channel = "shell"
 
-        return self._parents.get(channel, {})
+        if channel == "control":
+            return self._control_parent
+
+        return self._get_shell_context_var(self._shell_parent)
+
+    def _get_shell_context_var(self, var: ContextVar[T]) -> T:
+        """Lookup a ContextVar, falling back on the shell context
+
+        Allows for user-launched Threads to still resolve to the shell's main context
+
+        necessary for e.g. display from threads.
+        """
+        try:
+            return var.get()
+        except LookupError:
+            return self._shell_context[var]
 
     def send_response(
         self,
@@ -760,11 +818,19 @@ class Kernel(SingletonConfigurable):
         if self._do_exec_accepted_params["cell_id"]:
             do_execute_args["cell_id"] = cell_id
 
+        subshell_id = parent["header"].get("subshell_id")
+
         # Call do_execute with the appropriate arguments
         reply_content = self.do_execute(**do_execute_args)
 
         if inspect.isawaitable(reply_content):
             reply_content = await reply_content
+        else:
+            warnings.warn(
+                _AWAITABLE_MESSAGE.format(func_name="do_execute", target=self.do_execute),
+                PendingDeprecationWarning,
+                stacklevel=1,
+            )
 
         # Flush output before sending the reply.
         if sys.stdout is not None:
@@ -793,7 +859,8 @@ class Kernel(SingletonConfigurable):
         self.log.debug("%s", reply_msg)
 
         if not silent and reply_msg["content"]["status"] == "error" and stop_on_error:
-            self._abort_queues()
+            subshell_id = parent["header"].get("subshell_id")
+            self._abort_queues(subshell_id)
 
     def do_execute(
         self,
@@ -820,6 +887,12 @@ class Kernel(SingletonConfigurable):
         matches = self.do_complete(code, cursor_pos)
         if inspect.isawaitable(matches):
             matches = await matches
+        else:
+            warnings.warn(
+                _AWAITABLE_MESSAGE.format(func_name="do_complete", target=self.do_complete),
+                PendingDeprecationWarning,
+                stacklevel=1,
+            )
 
         matches = json_clean(matches)
         self.session.send(stream, "complete_reply", matches, parent, ident)
@@ -848,6 +921,12 @@ class Kernel(SingletonConfigurable):
         )
         if inspect.isawaitable(reply_content):
             reply_content = await reply_content
+        else:
+            warnings.warn(
+                _AWAITABLE_MESSAGE.format(func_name="do_inspect", target=self.do_inspect),
+                PendingDeprecationWarning,
+                stacklevel=1,
+            )
 
         # Before we send this object over, we scrub it for JSON usage
         reply_content = json_clean(reply_content)
@@ -867,6 +946,12 @@ class Kernel(SingletonConfigurable):
         reply_content = self.do_history(**content)
         if inspect.isawaitable(reply_content):
             reply_content = await reply_content
+        else:
+            warnings.warn(
+                _AWAITABLE_MESSAGE.format(func_name="do_history", target=self.do_history),
+                PendingDeprecationWarning,
+                stacklevel=1,
+            )
 
         reply_content = json_clean(reply_content)
         msg = self.session.send(stream, "history_reply", reply_content, parent, ident)
@@ -898,6 +983,14 @@ class Kernel(SingletonConfigurable):
 
     @property
     def kernel_info(self):
+        from .debugger import _is_debugpy_available
+
+        supported_features: list[str] = []
+        if self._supports_kernel_subshells:
+            supported_features.append("kernel subshells")
+        if _is_debugpy_available:
+            supported_features.append("debugger")
+
         return {
             "protocol_version": kernel_protocol_version,
             "implementation": self.implementation,
@@ -905,6 +998,7 @@ class Kernel(SingletonConfigurable):
             "language_info": self.language_info,
             "banner": self.banner,
             "help_links": self.help_links,
+            "supported_features": supported_features,
         }
 
     async def kernel_info_request(self, stream, ident, parent):
@@ -980,6 +1074,12 @@ class Kernel(SingletonConfigurable):
         content = self.do_shutdown(parent["content"]["restart"])
         if inspect.isawaitable(content):
             content = await content
+        else:
+            warnings.warn(
+                _AWAITABLE_MESSAGE.format(func_name="do_shutdown", target=self.do_shutdown),
+                PendingDeprecationWarning,
+                stacklevel=1,
+            )
         self.session.send(stream, "shutdown_reply", content, parent, ident=ident)
         # same content, but different msg_id for broadcasting on IOPub
         self._shutdown_message = self.session.msg("shutdown_reply", content, parent)
@@ -992,7 +1092,8 @@ class Kernel(SingletonConfigurable):
             control_io_loop.add_callback(control_io_loop.stop)
 
         self.log.debug("Stopping shell ioloop")
-        if self.shell_stream:
+        self.io_loop.add_callback(self.io_loop.stop)
+        if self.shell_stream and self.shell_stream.io_loop != self.io_loop:
             shell_io_loop = self.shell_stream.io_loop
             shell_io_loop.add_callback(shell_io_loop.stop)
 
@@ -1012,6 +1113,12 @@ class Kernel(SingletonConfigurable):
         reply_content = self.do_is_complete(code)
         if inspect.isawaitable(reply_content):
             reply_content = await reply_content
+        else:
+            warnings.warn(
+                _AWAITABLE_MESSAGE.format(func_name="do_is_complete", target=self.do_is_complete),
+                PendingDeprecationWarning,
+                stacklevel=1,
+            )
         reply_content = json_clean(reply_content)
         reply_msg = self.session.send(stream, "is_complete_reply", reply_content, parent, ident)
         self.log.debug("%s", reply_msg)
@@ -1028,6 +1135,14 @@ class Kernel(SingletonConfigurable):
         reply_content = self.do_debug_request(content)
         if inspect.isawaitable(reply_content):
             reply_content = await reply_content
+        else:
+            warnings.warn(
+                _AWAITABLE_MESSAGE.format(
+                    func_name="do_debug_request", target=self.do_debug_request
+                ),
+                PendingDeprecationWarning,
+                stacklevel=1,
+            )
         reply_content = json_clean(reply_content)
         reply_msg = self.session.send(stream, "debug_reply", reply_content, parent, ident)
         self.log.debug("%s", reply_msg)
@@ -1084,85 +1199,74 @@ class Kernel(SingletonConfigurable):
     async def do_debug_request(self, msg):
         raise NotImplementedError
 
-    # ---------------------------------------------------------------------------
-    # Engine methods (DEPRECATED)
-    # ---------------------------------------------------------------------------
+    async def create_subshell_request(self, socket, ident, parent) -> None:
+        """Handle a create subshell request.
+
+        .. versionadded:: 7
+        """
+        if not self.session:
+            return
+        if not self._supports_kernel_subshells:
+            self.log.error("Subshells are not supported by this kernel")
+            return
+
+        assert threading.current_thread().name == CONTROL_THREAD_NAME
+
+        # This should only be called in the control thread if it exists.
+        # Request is passed to shell channel thread to process.
+        control_socket = self.shell_channel_thread.manager.control_to_shell_channel.from_socket
+        control_socket.send_json({"type": "create"})
+        reply = control_socket.recv_json()
+        self.session.send(socket, "create_subshell_reply", reply, parent, ident)
+
+    async def delete_subshell_request(self, socket, ident, parent) -> None:
+        """Handle a delete subshell request.
+
+        .. versionadded:: 7
+        """
+        if not self.session:
+            return
+        if not self._supports_kernel_subshells:
+            self.log.error("KERNEL SUBSHELLS NOT SUPPORTED")
+            return
+
+        assert threading.current_thread().name == CONTROL_THREAD_NAME
 
-    async def apply_request(self, stream, ident, parent):  # pragma: no cover
-        """Handle an apply request."""
-        self.log.warning("apply_request is deprecated in kernel_base, moving to ipyparallel.")
         try:
             content = parent["content"]
-            bufs = parent["buffers"]
-            msg_id = parent["header"]["msg_id"]
+            subshell_id = content["subshell_id"]
         except Exception:
-            self.log.error("Got bad msg: %s", parent, exc_info=True)  # noqa: G201
+            self.log.error("Got bad msg from parent: %s", parent)
             return
 
-        md = self.init_metadata(parent)
+        # This should only be called in the control thread if it exists.
+        # Request is passed to shell channel thread to process.
+        control_socket = self.shell_channel_thread.manager.control_to_shell_channel.from_socket
+        control_socket.send_json({"type": "delete", "subshell_id": subshell_id})
+        reply = control_socket.recv_json()
 
-        reply_content, result_buf = self.do_apply(content, bufs, msg_id, md)
+        self.session.send(socket, "delete_subshell_reply", reply, parent, ident)
 
-        # flush i/o
-        if sys.stdout is not None:
-            sys.stdout.flush()
-        if sys.stderr is not None:
-            sys.stderr.flush()
+    async def list_subshell_request(self, socket, ident, parent) -> None:
+        """Handle a list subshell request.
 
-        md = self.finish_metadata(parent, md, reply_content)
+        .. versionadded:: 7
+        """
         if not self.session:
             return
-        self.session.send(
-            stream,
-            "apply_reply",
-            reply_content,
-            parent=parent,
-            ident=ident,
-            buffers=result_buf,
-            metadata=md,
-        )
-
-    def do_apply(self, content, bufs, msg_id, reply_metadata):
-        """DEPRECATED"""
-        raise NotImplementedError
-
-    # ---------------------------------------------------------------------------
-    # Control messages (DEPRECATED)
-    # ---------------------------------------------------------------------------
-
-    async def abort_request(self, stream, ident, parent):  # pragma: no cover
-        """abort a specific msg by id"""
-        self.log.warning(
-            "abort_request is deprecated in kernel_base. It is only part of IPython parallel"
-        )
-        msg_ids = parent["content"].get("msg_ids", None)
-        if isinstance(msg_ids, str):
-            msg_ids = [msg_ids]
-        if not msg_ids:
-            self._abort_queues()
-        for mid in msg_ids:
-            self.aborted.add(str(mid))
-
-        content = dict(status="ok")
-        if not self.session:
+        if not self._supports_kernel_subshells:
+            self.log.error("Subshells are not supported by this kernel")
             return
-        reply_msg = self.session.send(
-            stream, "abort_reply", content=content, parent=parent, ident=ident
-        )
-        self.log.debug("%s", reply_msg)
 
-    async def clear_request(self, stream, idents, parent):  # pragma: no cover
-        """Clear our namespace."""
-        self.log.warning(
-            "clear_request is deprecated in kernel_base. It is only part of IPython parallel"
-        )
-        content = self.do_clear()
-        if self.session:
-            self.session.send(stream, "clear_reply", ident=idents, parent=parent, content=content)
+        assert threading.current_thread().name == CONTROL_THREAD_NAME
 
-    def do_clear(self):
-        """DEPRECATED since 4.0.3"""
-        raise NotImplementedError
+        # This should only be called in the control thread if it exists.
+        # Request is passed to shell channel thread to process.
+        control_socket = self.shell_channel_thread.manager.control_to_shell_channel.from_socket
+        control_socket.send_json({"type": "list"})
+        reply = control_socket.recv_json()
+
+        self.session.send(socket, "list_subshell_reply", reply, parent, ident)
 
     # ---------------------------------------------------------------------------
     # Protected interface
@@ -1176,15 +1280,37 @@ class Kernel(SingletonConfigurable):
 
     _aborting = Bool(False)
 
-    def _abort_queues(self):
+    def _post_dummy_stop_aborting_message(self, subshell_id: str | None) -> None:
+        """Post a dummy message to the correct subshell that when handled will unset
+        the _aborting flag.
+        """
+        subshell_manager = self.shell_channel_thread.manager
+        socket = subshell_manager.get_shell_channel_to_subshell_socket(subshell_id)
+        assert socket is not None
+
+        msg = b"stop aborting"  # Magic string for dummy message.
+        socket.send(msg, copy=False)
+
+    def _abort_queues(self, subshell_id: str | None = None):
         # while this flag is true,
         # execute requests will be aborted
-        self._aborting = True
+
+        if subshell_id is None:
+            self._aborting = True
+        else:
+            self.shell_channel_thread.manager.set_subshell_aborting(subshell_id, True)
         self.log.info("Aborting queue")
 
+        if self.shell_channel_thread:
+            # Only really need to do this if there are messages already queued
+            self.shell_channel_thread.io_loop.add_callback(
+                self._post_dummy_stop_aborting_message, subshell_id
+            )
+            return
+
         # flush streams, so all currently waiting messages
         # are added to the queue
-        if self.shell_stream:
+        if self.shell_stream and not self._supports_kernel_subshells:
             self.shell_stream.flush()
 
         # Callback to signal that we are done aborting
@@ -1193,15 +1319,10 @@ class Kernel(SingletonConfigurable):
             self.log.info("Finishing abort")
             self._aborting = False
 
-        # put the stop-aborting event on the message queue
-        # so that all messages already waiting in the queue are aborted
-        # before we reset the flag
-        schedule_stop_aborting = partial(self.schedule_dispatch, stop_aborting)
-
         if self.stop_on_error_timeout:
             # if we have a delay, give messages this long to arrive on the queue
             # before we stop aborting requests
-            self.io_loop.call_later(self.stop_on_error_timeout, schedule_stop_aborting)
+            self.io_loop.call_later(self.stop_on_error_timeout, stop_aborting)
             # If we have an eventloop, it may interfere with the call_later above.
             # If the loop has a _schedule_exit method, we call that so the loop exits
             # after stop_on_error_timeout, returning to the main io_loop and letting
@@ -1209,7 +1330,7 @@ class Kernel(SingletonConfigurable):
             if self.eventloop is not None and hasattr(self.eventloop, "_schedule_exit"):
                 self.eventloop._schedule_exit(self.stop_on_error_timeout + 0.01)
         else:
-            schedule_stop_aborting()
+            self.io_loop.add_callback(stop_aborting)
 
     def _send_abort_reply(self, stream, msg, idents):
         """Send a reply to an aborted request"""
@@ -1257,7 +1378,7 @@ class Kernel(SingletonConfigurable):
             )
         return self._input_request(
             prompt,
-            self._parent_ident["shell"],
+            self._get_shell_context_var(self._shell_parent_ident),
             self.get_parent("shell"),
             password=True,
         )
@@ -1274,7 +1395,7 @@ class Kernel(SingletonConfigurable):
             raise StdinNotImplementedError(msg)
         return self._input_request(
             str(prompt),
-            self._parent_ident["shell"],
+            self._get_shell_context_var(self._shell_parent_ident),
             self.get_parent("shell"),
             password=False,
         )
@@ -1412,3 +1533,7 @@ class Kernel(SingletonConfigurable):
                 self.log.debug("%s", self._shutdown_message)
             if self.control_stream:
                 self.control_stream.flush(zmq.POLLOUT)
+
+    @property
+    def _supports_kernel_subshells(self):
+        return self.shell_channel_thread is not None
